from dcicutils.codebuild_utils import CodeBuildUtils
from .helpers.confchecks import *


@check_function()
def codebuild_status(connection, **kwargs):
    """ Returns list of available CodeBuild projects """
    check = CheckResult(connection, 'codebuild_status')
    client = CodeBuildUtils()
    projects = client.list_projects()
    check.status = 'PASS'
    check.summary = 'See full output for info on valid build names'
    check.full_output = projects
    return check


@check_function(build_name=None)
def trigger_codebuild_run(connection, **kwargs):
    """ This checks triggers a run of the CodeBuild pipeline
        Usually, there is only 1 CodeBuild pipeline per account - this function
        will work out of the box where there is only 1, and will require kwarg
        if there are multiple.
    """
    check = CheckResult(connection, 'trigger_codebuild_run')
    client = CodeBuildUtils()
    projects = client.list_projects()
    build_name = kwargs.get('build_name', None) or (projects[0] if len(projects) == 1 else None)
    if build_name:
        resp = client.run_project_build(project_name=build_name)
        check.full_output = resp
        check.status = 'PASS'
        check.summary = f'Triggered build {build_name}'
    else:
        check.status = 'FAIL'
        check.summary = f'Cannot resolve which build you want: {projects}'
    return check


@check_function(build_name=None, branch=None, image_repo_name=None, image_tag=None, build_path=None)
def trigger_codebuild_run_with_overrides(connection, **kwargs):
    """ Very similar to the above function but takes some additional arguments it can work with to change
        the branch, image tag and build path if needed.

        You MUST pass a branch to this at a minimum, but other arguments are not necessary
    """
    check = CheckResult(connection, 'trigger_codebuild_run')
    client = CodeBuildUtils()
    projects = client.list_projects()
    project = kwargs.get('build_name', None) or (projects[0] if len(projects) == 1 else None)
    if not project:
        check.status = 'FAIL'
        check.summary = f'Cannot resolve which build you want: {projects}'
        return check
    branch = kwargs.get('branch', None)
    image_tag = kwargs.get('image_tag', None)
    build_path = kwargs.get('build_path', None)
    image_repo_name = kwargs.get('image_repo_name', None)
    env_overrides = {}
    if image_repo_name:
        env_overrides['IMAGE_REPO_NAME'] = image_repo_name
    if image_tag:
        env_overrides['IMAGE_TAG'] = image_tag
    if build_path:
        env_overrides['BUILD_PATH'] = build_path
    resp = client.run_project_build_with_overrides(project_name=project, branch=branch, env_overrides=env_overrides)
    check.full_output = resp
    check.status = 'PASS'
    check.summary = f'Triggered build {project} from branch {branch} with overrides {env_overrides}'
    return check



from datetime import datetime, timedelta
import pytz
from dcicutils import ff_utils
from dcicutils.misc_utils import future_datetime
from typing import Optional, Tuple
from ...app import app
from .datetime_utils import (
    convert_iso_datetime_string_to_datetime as normalize_portal_datetime,
    convert_utc_datetime_to_utc_datetime_string as datetime_to_string
)


_PORTAL_ACCESS_KEY_NAME = "access_key_foursight"
_PORTAL_ACCESS_KEY_USER_EMAIL = "foursight.app@gmail.com"
_PORTAL_ACCESS_KEY_EXPIRES_SOON_WARNING_DAYS = 12


def get_portal_access_key_info(env: str,
                               logged_in: bool = False,
                               test_mode_access_key_simulate_error: bool = False,
                               test_mode_access_key_expiration_warning_days: int = 0) -> dict:
    try:
        now = datetime.now(pytz.utc)
        connection = app.core.init_connection(env)
        connection_keys = connection.ff_keys
        key = connection_keys.get("key")
        if key and not logged_in:
            # This logged_in flag is passed from caller (react_api.reactapi_portal_access_key)
            # and is False iff the user is NOT logged in; so if logged in then we do NOT
            # obfuscate the access key (ID) but if we are NOT logged in we do obfuscate it.
            # And of course we do not return any part of the secret at all, ever.
            key = "********"
        server = connection_keys.get("server")
        access_key_info = {"key": key, "secret": "********", "server": server, "logged_in": logged_in}
        access_key_create_date, access_key_expires_date, access_key_expires_exception = \
            _get_portal_access_key_expires_date(connection_keys)
        if access_key_expires_date:
            access_key_info["created_at"] = datetime_to_string(access_key_create_date)
            if access_key_expires_date == datetime.max:
                access_key_info["expires_at"] = None
                access_key_info["expired"] = False
                access_key_info["invalid"] = False
                access_key_info["expires_soon"] = False
            else:
                access_key_info["expires_at"] = datetime_to_string(access_key_expires_date)
                access_key_info["expired"] = now >= access_key_expires_date
                access_key_info["invalid"] = access_key_info["expired"]
                # Note that these "test_mode_xyz" variables are for testing only and are set
                # via cookies and if used must be manually set, e.g. via Chrome Developer Tools.
                if test_mode_access_key_expiration_warning_days > 0:
                    expires_soon_days = test_mode_access_key_expiration_warning_days
                else:
                    expires_soon_days = _PORTAL_ACCESS_KEY_EXPIRES_SOON_WARNING_DAYS
                access_key_info["expires_soon"] = _is_datetime_within_future_ndays(access_key_expires_date,
                                                                                   ndays=expires_soon_days,
                                                                                   now=now)
            if test_mode_access_key_simulate_error:
                access_key_info["exception"] = "test_mode_access_key_simulate_error"
                access_key_info["invalid"] = True
        else:
            # If we don't get an expires date at all, then assume there is a problem.
            access_key_info["exception"] = str(access_key_expires_exception)
            e = str(access_key_info["exception"]).lower()
            if "credenti" in e or "expir" in e:
                access_key_info["probably_expired"] = True
            access_key_info["invalid"] = True
        access_key_info["timestamp"] = datetime_to_string(now)
        return access_key_info
    except Exception:
        return {}


def _get_portal_access_key_expires_date(keys: dict) -> Tuple[datetime, Optional[datetime], Optional[Exception]]:
    try:
        query = f"/search/?type=AccessKey&description={_PORTAL_ACCESS_KEY_NAME}&sort=-date_created"
        access_key = ff_utils.search_metadata(query, key=keys)[0]
        access_key_create_date = normalize_portal_datetime(access_key["date_created"])
        access_key_expires_date = access_key.get("expiration_date")
        if access_key_expires_date:
            access_key_expires_date = normalize_portal_datetime(access_key_expires_date)
        else:
            # There may or may not be an expiration date (e.g. for fourfront);
            # if not then make it the max date which is 9999-12-31 23:59:59.999999.
            access_key_expires_date = datetime.max
        return (access_key_create_date, access_key_expires_date, None)
    except Exception as e:
        return (None, None, e)


def _is_datetime_within_future_ndays(d: datetime, ndays: int, now: Optional[datetime] = None) -> bool:
    """
    Returns True iff the given datatime is within ndays IN THE FUTURE of the CURRENT
    datetime. I.e. if the current datetime plus ndays is greater-than-or-equal to
    the given datetime the returns True; otherwise returns False.
    """
    if not now:
        now = datetime.now(pytz.utc)
    return d <= future_datetime(now=now, days=ndays)



"""
Generate gitignored .chalice/config.json for deploy and then run deploy.
Takes on parameter for now: stage (either "dev" or "prod")
"""
from dcicutils.misc_utils import PRINT
import os
import sys
import argparse
import json
import subprocess

from dcicutils.misc_utils import as_seconds, ignored


# TODO: Move to dcicutils.command_utils.
def subprocess_call(command: list, verbose: bool = False, **kwargs) -> int:
    """
    Executes the given command (list of arguments, the main command as the zeroth)
    in a sub-process and returns the return code from the command execution.
    Prints (to stdout) before/after execution messages if ``verbose`` is True.
    If general (not command execution related) exception occurs raises that exception.

    :param command: Command as list with zeroth item being the main command.
    :param verbose: If True prints (to stdout) before/after execution messages.
    :return: Return code from the command execution.
    """
    if verbose:
        command_string = " ".join(command)
    try:
        if verbose:
            PRINT(f"Executing command in sub-process: {command_string}")
        return_code = subprocess.check_call(command, **kwargs)
        if verbose:
            PRINT(f"Done executing (return-code: {return_code}) command in sub-process: {command_string}")
    except subprocess.CalledProcessError as command_exception:
        return_code = command_exception.returncode
        if verbose:
            PRINT(f"Error executing (return-code: {return_code}) command in sub-process: {command_string}")
    except Exception as general_exception:
        if verbose:
            PRINT(f"Exception executing command ({command_string}) in sub-process: {general_exception}")
        raise general_exception
    return return_code


class Deploy(object):

    DEFAULT_LAMBDA_TIMEOUT = as_seconds(minutes=15)

    CONFIG_BASE = {
      "stages": {
        "dev": {
          "api_gateway_stage": "api",
          "autogen_policy": False,
          "lambda_memory_size": 512,
          "lambda_timeout": DEFAULT_LAMBDA_TIMEOUT,  # 15 mins in seconds
          "environment_variables": {
              "chalice_stage": "dev"
          }
        },
        "prod": {
          "api_gateway_stage": "api",
          "autogen_policy": False,
          "lambda_memory_size": 512,
          "lambda_timeout": DEFAULT_LAMBDA_TIMEOUT,  # 15 mins in seconds
          "environment_variables": {
              "chalice_stage": "prod"
          }
        }
      },
      "version": "2.0",
      "app_name": "foursight-cgap",
      "layers": [
          "arn:aws:lambda:us-east-1:553035198032:layer:git:11"  # required for Deployment Checks - Will 5/20/2020
      ]
    }

    config_dir = os.path.dirname(__file__)

    @classmethod
    def get_config_filepath(cls):
        return os.path.join(cls.config_dir, '.chalice/config.json')

    @classmethod
    # dmichaels/2022-07-22/C4-826:
    # Added identity and stack_name args for environment variables in the Foursight CloudFormation template.
    # The identity is the global application configuration (GAC) name from which we read other pertinent
    # secrets/values (e.g. CLIENT_ID/SECRET) at runtime. The stack_name is the Foursight CloudFormation
    # stack name which provides a foothold to get other stack related values at runtime (e.g. the name
    # of the CheckerRunner lambda function generated by chalice at foursight provision time, which we
    # get by looking for the AWS lambda function using the regex "{stack_name}-CheckRunner-.*").
    def build_config(cls, stage, identity=None, stack_name=None,
                     trial_creds=None, trial_global_env_bucket=False, global_env_bucket=None,
                     security_group_ids=None, subnet_ids=None, check_runner=None,
                     lambda_timeout=DEFAULT_LAMBDA_TIMEOUT, is_foursight_fourfront=False,
                     is_foursight_smaht=False):
        """ Builds the chalice config json file. See: https://aws.github.io/chalice/topics/configfile"""
        # dmichaels/2022-07-22/C4-826:
        # Removed value from the Foursight CloudFormation template; get from GAC/etc at runtime.
        ignored(stage)
        if trial_creds:
            # key to decrypt access key
            # s3_enc_secret = trial_creds['S3_ENCRYPT_KEY']
            # client_id = trial_creds['CLIENT_ID']
            # client_secret = trial_creds['CLIENT_SECRET']
            dev_secret = None
            # es_host = trial_creds['ES_HOST']
            # env_name = trial_creds['ENV_NAME']
            # rds_name = trial_creds['RDS_NAME']
            # s3_key_id = trial_creds.get('S3_ENCRYPT_KEY_ID')
            # if not (s3_enc_secret and client_id and client_secret and es_host and rds_name):
            #     PRINT(''.join(['ERROR. You are missing one more more environment',
            #                    'variables needed to deploy the Foursight trial. Need:\n',
            #                    'S3_ENCRYPT_KEY, CLIENT_ID, CLIENT_SECRET, ES_HOST, RDS_NAME in trial_creds dict.'])
            #           )
            #     sys.exit()
        else:
            # s3_enc_secret = os.environ.get("S3_ENCRYPT_KEY")
            # client_id = os.environ.get("CLIENT_ID")
            # client_secret = os.environ.get("CLIENT_SECRET")
            dev_secret = os.environ.get("DEV_SECRET")
            # es_host = None  # not previously passed to config
            # env_name = None
            # s3_key_id = None  # not supported in legacy
            # if not (s3_enc_secret and client_id and client_secret and dev_secret):
            #     PRINT(''.join(['ERROR. You are missing one more more environment ',
            #                    'variables needed to deploy Foursight.\n',
            #                    'Need: S3_ENCRYPT_KEY, CLIENT_ID, CLIENT_SECRET, DEV_SECRET.'])
            #           )
            #     sys.exit()
        for curr_stage_name in ['dev', 'prod']:
            curr_stage = cls.CONFIG_BASE['stages'][curr_stage_name]
            curr_stage_environ = curr_stage['environment_variables']

            # dmichaels/2022-07-22/C4-826:
            # Added these to the Foursight CloudFormation template; get other values from GAC/etc at runtime.
            curr_stage_environ['IDENTITY'] = identity
            curr_stage_environ['STACK_NAME'] = stack_name
            # dmichaels/2022-07-22/C4-826:
            # Removed these from the Foursight CloudFormation template; get from GAC/etc at runtime.
            # curr_stage_environ['S3_ENCRYPT_KEY'] = s3_enc_secret
            # curr_stage_environ['CLIENT_ID'] = client_id
            # curr_stage_environ['CLIENT_SECRET'] = client_secret
            # if rds_name:
            #     curr_stage_environ['RDS_NAME'] = rds_name
            if dev_secret:  # still pass in main account, ignored in alpha infra - Will Aug 24 2021
                curr_stage_environ['DEV_SECRET'] = dev_secret
            # dmichaels/2022-07-22/C4-826:
            # Removed these from the Foursight CloudFormation template; get from GAC/etc at runtime.
            # if env_name:
            #     curr_stage_environ['ENV_NAME'] = env_name
            # if es_host:
            #     curr_stage_environ['ES_HOST'] = es_host
            # if s3_key_id:
            #     curr_stage_environ['S3_ENCRYPT_KEY_ID'] = s3_key_id
            if trial_global_env_bucket:
                # in the trial account setup, use a shorter timeout
                curr_stage['lambda_timeout'] = lambda_timeout
                if not global_env_bucket:
                    global_bucket_env_from_environ = os.environ.get('GLOBAL_BUCKET_ENV')
                    global_env_bucket_from_environ = os.environ.get('GLOBAL_ENV_BUCKET')
                    if (global_bucket_env_from_environ
                            and global_env_bucket_from_environ
                            and global_bucket_env_from_environ != global_env_bucket_from_environ):
                        PRINT('ERROR. GLOBAL_BUCKET_ENV and GLOBAL_ENV_BUCKET are both set, but inconsistently.')
                        sys.exit(1)
                    global_env_bucket = global_bucket_env_from_environ or global_env_bucket_from_environ
                if global_env_bucket:
                    curr_stage_environ['GLOBAL_BUCKET_ENV'] = global_env_bucket  # legacy compatibility
                    curr_stage_environ['GLOBAL_ENV_BUCKET'] = global_env_bucket
                else:
                    PRINT('ERROR. GLOBAL_ENV_BUCKET must be set or global_env_bucket= must be passed'
                          ' when building the trial config.')
                    sys.exit(1)
            if security_group_ids:
                curr_stage['security_group_ids'] = security_group_ids
            if subnet_ids:
                curr_stage['subnet_ids'] = subnet_ids
            if check_runner:
                curr_stage_environ['CHECK_RUNNER'] = check_runner

        filename = cls.get_config_filepath()
        PRINT(''.join(['Writing: ', filename]))
        with open(filename, 'w') as config_file:
            config_file.write(json.dumps(cls.CONFIG_BASE))
        PRINT(''.join(['Done writing: ', filename]))
        # Export poetry into requirements.
        # Note that depending on is_foursight_fourfront we include either foursight_fourfront or foursight_cgap;
        # these names are poetry group names defined in 4dn-cloud-infra/pyproject.toml. dmichaels/2022-11-01.
        if is_foursight_fourfront:
            subprocess_call(
                ['poetry', 'export', '-f', 'requirements.txt', '--without-hashes', '--with', 'foursight_fourfront', '-o', 'requirements.txt'], verbose=True)
        elif is_foursight_smaht:
            subprocess_call(
                ['poetry', 'export', '-f', 'requirements.txt', '--without-hashes', '--with', 'foursight_smaht',
                 '-o', 'requirements.txt'], verbose=True)
        else:
            subprocess_call(
                ['poetry', 'export', '-f', 'requirements.txt', '--without-hashes', '--with', 'foursight_cgap', '-o', 'requirements.txt'], verbose=True)

    @classmethod
    def build_config_and_deploy(cls, stage):
        cls.build_config(stage)
        # actually deploy
        subprocess_call(['chalice', 'deploy', '--stage', stage], verbose=True)

    @classmethod
    # dmichaels/2022-07-22/C4-826:
    # Added identity arg for the Foursight CloudFormation template; get other values from GAC/etc at runtime.
    def build_config_and_package(cls, args, identity=None, stack_name=None, trial_creds=None, global_env_bucket=None,
                                 security_ids=None, subnet_ids=None, check_runner=None,
                                 lambda_timeout=DEFAULT_LAMBDA_TIMEOUT,
                                 # These next args are preferred over passing 'args'.
                                 merge_template=None, output_file=None, stage=None, trial=None,
                                 ):
        """ Builds a config with a special case for the trial account. For the trial account, expects a dictionary of
            environment variables, a list of security group ids, and a list of subnet ids. Finally, packages as a
            Cloudformation template."""

        # Determine if we're packaging Foursight-CGAP or Foursight-Fourfront, based on
        # the provision stack name setup in 4dn-cloud-infra/stack.py. This is used to
        # conditionally include the appropriate library (foursight-cgap or foursight)
        # in the Chalice package. dmichaels/2022-11-01.
        # Added smaht to this to minimally perturb existing structure, but should be
        # refactored at a later time to clean up and use EnvUtils
        _4DN_CLOUD_INFRA_FOURSIGHT_FOURFRONT_PROVISION_TARGETS = ["foursight-development", "foursight-production"]
        _4DN_CLOUD_INFRA_FOURSIGHT_SMAHT_PROVISION_TARGETS = ["foursight-smaht"]
        is_foursight_fourfront = False
        is_foursight_smaht = False
        if args.stack in _4DN_CLOUD_INFRA_FOURSIGHT_FOURFRONT_PROVISION_TARGETS:
            is_foursight_fourfront = True
        elif args.stack in _4DN_CLOUD_INFRA_FOURSIGHT_SMAHT_PROVISION_TARGETS:
            is_foursight_smaht = True

        # For compatibility during transition, we allow these argument to be passed in lieu of args.
        if merge_template is None:
            merge_template = args.merge_template
        if output_file is None:
            output_file = args.output_file
        if stage is None:
            stage = args.stage
        if trial is None:
            trial = args.trial

        if trial:
            # Note we no longer need check_runner (previously checked here) as it is determined
            # dynamically at runtime (in identity.set_check_runner_lambda_environment_variable);
            # should have been removed before. dmichaels/2022-11-01.
            if trial_creds and security_ids and subnet_ids:
                # dmichaels/2022-07-22/C4-826:
                # Added identity arg for the Foursight CloudFormation template.
                cls.build_config(stage, identity=identity, stack_name=stack_name, trial_creds=trial_creds,
                                 trial_global_env_bucket=True,
                                 global_env_bucket=global_env_bucket, lambda_timeout=lambda_timeout,
                                 security_group_ids=security_ids, subnet_ids=subnet_ids, check_runner=check_runner,
                                 is_foursight_fourfront=is_foursight_fourfront, is_foursight_smaht=is_foursight_smaht)
            else:
                raise Exception('Build config requires trial_creds, sg id, and subnet ids to run in trial account')
        else:
            cls.build_config(stage=stage)
        # actually package cloudformation templates
        # add --single-file ?
        flags = ['--stage', stage, '--pkg-format', 'cloudformation', '--template-format', 'yaml']
        if merge_template:
            flags.extend(['--merge-template', merge_template])
        PRINT(f"Starting chalice package: {output_file}")
        subprocess_call(['chalice', 'package', *flags, output_file], verbose=True)
        PRINT(f"Finished chalice package: {output_file}")

        # dmichaels/2022-12-16/C4-???:
        # We reached the maximum size of a Chalice package to deploy in AWS (220,476,770 bytes)
        # Prune out of it unnecessary modules, which look like tests and/or examples.
        # This prune process is done by this bash script which lives in 4dn-cloud-infra:
        # scripts/prune_chalice_package.sh.
        prune_chalice_package_script_file = os.path.join(os.getcwd(), "scripts/prune_chalice_package.sh")
        if os.path.exists(prune_chalice_package_script_file):
            chalice_package_file = os.path.join(output_file, "deployment.zip")
            if not os.path.exists(chalice_package_file):
                PRINT(f"WARNING: Chalice package file not found: {chalice_package_file}")
                PRINT(f"WARNING: Not running chalice prune script.")
            else:
                PRINT(f"Found chalice prune script: {prune_chalice_package_script_file}")
                PRINT(f"Starting chalice package prune: {chalice_package_file}")
                subprocess_call([prune_chalice_package_script_file, chalice_package_file], verbose=True)
                PRINT(f"Finished chalice package prune: {chalice_package_file}")
        else:
            PRINT(f"No chalice prune script found.")


def main():
    parser = argparse.ArgumentParser('chalice_deploy')
    parser.add_argument(
        "stage",
        type=str,
        choices=['dev', 'prod'],
        help="chalice deployment stage. Must be one of 'prod' or 'dev'")
    args = parser.parse_args()
    Deploy.build_config_and_deploy(stage=args.stage)


if __name__ == '__main__':
    main()



import base64
import json
from typing import Union


def base64_encode(value: Union[str, dict, list]) -> str:
    if isinstance(value, dict) or isinstance(value, list):
        value = json.dumps(value)
    elif not isinstance(value, str):
        raise ValueError("Invalid argument to base64_encode; must be string, dict, or bytes.")
    return base64_encode_to_bytes(value).decode("utf-8")


def base64_decode(value: Union[str, bytes]) -> str:
    return base64_decode_to_bytes(value).decode("utf-8")


def base64_decode_to_json(value: str) -> Union[dict, list, str, int, float, bool]:
    return json.loads(base64_decode(value))


def base64_encode_to_bytes(string_or_bytes: Union[str, bytes]) -> bytes:
    if isinstance(string_or_bytes, str):
        return base64.b64encode(string_to_bytes(string_or_bytes))
    elif isinstance(string_or_bytes, bytes):
        return base64.b64encode(string_or_bytes)
    else:
        raise ValueError("Invalid argument to base64_encode_to_bytes; must be string or bytes.")


def base64_decode_to_bytes(string_or_bytes: Union[str, bytes]) -> bytes:
    if isinstance(string_or_bytes, str):
        return base64.b64decode(string_to_bytes(string_or_bytes))
    elif isinstance(string_or_bytes, bytes):
        return base64.b64decode(string_or_bytes)
    else:
        raise ValueError("Invalid argument to base64_decode_to_bytes; must be string or bytes.")


def string_to_bytes(value: str) -> bytes:
    if isinstance(value, str):
        return value.encode("utf-8")
    else:
        raise ValueError("Invalid argument to string_to_bytes; must be string.")


def bytes_to_string(value: bytes) -> str:
    if isinstance(value, bytes):
        return value.decode("utf-8") if isinstance(value, bytes) else ""
    else:
        raise ValueError("Invalid argument to bytes_to_string; must be bytes.")



# The primary/initial purpose of this was to be able to use the SQS_URL environment variable to refer
# to a locally running ersatz version of SQS via an emulator like localstack (https://localstack.cloud).

import boto3
import os


def boto_sqs_client(**kwargs):
    """
    Creates and returns a boto3 sqs client object. If the SQS_URL environment variable is set then it
    will use that value as the endpoint_url for the boto3 sqs client, unless an explicit endpoint_url
    was passed in (via kwargs, per boto3.client convention) in which case that value will be used.
    """
    sqs_url = kwargs.get("endpoint_url") or os.environ.get("SQS_URL")
    return boto3.client("sqs", endpoint_url=sqs_url, **kwargs) if sqs_url else boto3.client("sqs", **kwargs)


def boto_sqs_resource(**kwargs):
    """
    Creates and returns a boto3 sqs resource object. If the SQS_URL environment variable is set then it
    will use that value as the endpoint_url for the boto3 sqs resource, unless an explicit endpoint_url
    was passed in (via kwargs, per boto3.resource convention) in which case that value will be used.
    """
    sqs_url = kwargs.get("endpoint_url") or os.environ.get("SQS_URL")
    return boto3.resource("sqs", endpoint_url=sqs_url, **kwargs) if sqs_url else boto3.resource("sqs", **kwargs)



import datetime
import dateutil.parser as dateutil_parser
import pytz
from typing import Optional, Union


EPOCH = datetime.datetime.utcfromtimestamp(0)  # I.e.: 1970-01-01 00:00:00 UTC


def _convert_utc_datetime_to_datetime_string(t: Union[datetime.datetime, str],
                                             tzname: Optional[str] = None) -> Optional[str]:
    """
    Converts the given datetime object OR string, which is ASSUMED to by in the UTC timezone,
    into a datetime string in the given/named timezone, and returns its value in a form that looks
    like 2022-08-22 13:25:34 EDT. If the argument is a string it is ASSUMED to have a value which
    looks like 2022-08-22T14:24:49.000+0000; this is (for example) the format we get from AWS via
    boto3 (e.g. for a lambda last-modified value) or from values in ElasticSearch.

    :param t: A datetime object or string value ASSUMED to be in the UTC timezone.
    :param tzname: A timezone name (string); default to UTC if unspecified.
    :return: A datetime string in the given timezone formatted like: 2022-08-22 13:25:34 EDT
    """
    def make_utc_aware_datetime(t: datetime) -> datetime:
        return t.replace(tzinfo=pytz.UTC)
    try:
        if isinstance(t, str):
            #
            # Can sometimes get dates (from user ElasticSearch index)
            # which look like this: 2019-06-20T00:00:00.0000000+00:00
            # i.e. with 7-digits for ms which does not parse (up to 6).
            # Was doing this hack below but found that the Python
            # dateutil.parser is more forgiving so using that now.
            #
            # if ".0000000" in t:
            #     t = t.replace(".0000000", ".000000")
            # t = datetime.datetime.strptime(t, "%Y-%m-%dT%H:%M:%S.%f%z")
            #
            t = dateutil_parser.parse(t)
        tz = pytz.timezone(tzname) if tzname else pytz.UTC
        t = make_utc_aware_datetime(t).astimezone(tz)
        return t.strftime("%Y-%m-%d %H:%M:%S %Z")
    except Exception as e:
        return None


def convert_utc_datetime_to_utc_datetime_string(t: Union[datetime.datetime, str]) -> Optional[str]:
    """
    Same as _convert_utc_datetime_to_datetime_string (above) but explicitly for the UTC timezone.
    """
    return _convert_utc_datetime_to_datetime_string(t)


def convert_time_t_to_datetime_string(time_t: int, tzname: Optional[str] = None) -> Optional[str]:
    """
    Converts the given "epoch" time (seconds since 1970-01-01T00:00:00Z)
    integer value to a datetime string in the given/named timezone,
    and returns its value in a form that looks like 2022-08-22 13:25:34 EDT.

    :param time_t: Epoch time value (i.e. seconds since 1970-01-01T00:00:00Z)
    :param tzname: A timezone name (string); default to UTC if unspecified.
    :return: A datetime string in the given timezone formatted like: 2022-08-22 13:25:34 EDT
    """
    return _convert_utc_datetime_to_datetime_string(convert_time_t_to_datetime(time_t), tzname)


def convert_time_t_to_utc_datetime_string(time_t: int) -> Optional[str]:
    """
    Same as convert_time_t_to_datetime_string (above) but explicitly for the UTC timezone.
    """
    return convert_time_t_to_datetime_string(time_t)


def convert_iso_datetime_string_to_datetime(value: str) -> Optional[datetime.datetime]:
    """
    Returns a datetime object, in the UTC timezone, for the given (assumed)
    ISO 8601 conforming datetime string; if an error occurs (e.g. if the
    given value is not ISO 8601 conforming) then returns None.
    """
    try:
        return datetime.datetime.fromisoformat(value).astimezone(pytz.utc)
    except Exception:
        return None


def convert_time_t_to_datetime(time_t: int) -> datetime.datetime:
    """
    Converts the given "epoch" time (seconds since 1970-01-01T00:00:00Z)
    to a datetime and returns its value.
    """
    return datetime.datetime.fromtimestamp(time_t, tz=pytz.UTC)


def convert_datetime_to_time_t(value: datetime.datetime) -> int:
    """
    Converts the given datatime value to an "epoch"
    time (seconds since 1970-01-01T00:00:00Z) and returns its value.
    """
    return int(value.timestamp())


def convert_utc_datetime_to_cookie_expires_format(t) -> str:
    """
    Converts the given UTC datetime object or string format suitable
    for use by a cookie expires date/time value.
    """
    return t.strftime("%a, %d %b %Y %H:%M:%S GMT")


def convert_uptime_to_datetime(uptime: str, relative_to: datetime = None) -> Optional[datetime.datetime]:
    """
    Converts the given duration string which (happens to be) from the Portal health endpoint
    into its equivalent datetime. Format of the given upatime looks something like this:

      1 week, 2 days, 3 hours, 4 minutes, 5.67 seconds

    If the given uptime is not parsable then returns None, otherwise returns the datetime corresponding
    to this uptime, relative to now, by default, or to the relative_to datetime argument if given.
    We THINK it's right to interpret the given uptime relative to UTC (TODO).
    """
    def normalize_spaces(s: str) -> str:
        return " ".join(s.split())

    if not uptime:
        return None
    try:
        minutes_per_hour = 60
        minutes_per_day = minutes_per_hour * 24
        minutes_per_week = minutes_per_day * 7
        minutes = 0
        seconds = 0
        uptime = normalize_spaces(uptime)
        for item in uptime.split(","):
            item = item.strip()
            if item:
                item = item.split(" ")
                if len(item) == 2:
                    unit = item[1].lower()
                    value = float(item[0])
                    if unit.startswith("week"):
                        minutes += minutes_per_week * value
                    elif unit.startswith("day"):
                        minutes += minutes_per_day * value
                    elif unit.startswith("hour"):
                        minutes += minutes_per_hour * value
                    elif unit.startswith("minute"):
                        minutes += value
                    elif unit.startswith("second"):
                        seconds += value
        t = relative_to if relative_to else datetime.datetime.now(datetime.timezone.utc)
        return t + datetime.timedelta(minutes=-minutes, seconds=-seconds)
    except Exception:
        pass
    return None



from chalice import Response
from dcicutils.misc_utils import ignored
from ...app import app
from .misc_utils import get_request_arg, get_request_args, get_request_body
from .react_route_decorator import route, route_root


class ReactRoutes:

    def __init__(self):
        super(ReactRoutes, self).__init__()

    # ----------------------------------------------------------------------------------------------
    # Foursight React API routes UNPROTECTED by authorization/authentication.
    # ----------------------------------------------------------------------------------------------

    @route("/{env}/auth0_config", authorize=False)
    def reactapi_route_auth0_config(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns Auth0 configuration for authentication (from Portal).
        Note that this in an UNPROTECTED route.
        """
        ignored(env)
        return ReactRoutes.reactapi_route_auth0_config_noenv()

    @route("/auth0_config", authorize=False)
    def reactapi_route_auth0_config_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns Auth0 configuration for authentication (from Portal).
        Note that this in an UNPROTECTED route.
        """
        return app.core.reactapi_auth0_config(app.current_request.to_dict())

    @route("/cognito_config", authorize=False)
    def reactapi_route_cognito_config() -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns AWS Cognito configuration for authentication; from environment variables or Secrets Manager;
        getting from one or the other happens in cognito.get_cognito_oauth_config.
        Note that this in an UNPROTECTED route.
        """
        return app.core.reactapi_cognito_config(app.current_request.to_dict())

    @route("/cognito/callback", authorize=False)
    def reactapi_route_cognito_callback() -> Response:  # noqa: implicit @staticmethod via @route
        # API version of reactapi_route_cognito_callback.
        """
        Secondary AWS Cognito OAuth callback; called from our primary frontend callback which is
        redirected to from Cognito so it can pick up the ouath_pkce_key (sic) which is written to
        browser session storage by the React authentication kickoff code (Amplify.federatedSignIn).
        Note that this in an UNPROTECTED route.
        """
        return app.core.reactapi_cognito_callback(app.current_request.to_dict())

    @route("/{env}/logout", authorize=False)
    def reactapi_route_logout(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        Logs out the user. Sends back specific message if already logged out, or not.
        The env is not strictly required for logout, since we logout from all environments,
        but it is is useful for the redirect back, and also just for consistency/completeness.
        """
        return app.core.reactapi_logout(app.current_request.to_dict(), env)

    @route("/logout", authorize=False)
    def reactapi_route_logout_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        No-env version of above /{env}/logout route.
        """
        return app.core.reactapi_logout(app.current_request.to_dict(), app.core.get_default_env())

    @route("/{env}/header", authorize=False)
    def reactapi_route_header(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        Returns minimal data for React UI to get up and running.
        """
        return app.core.reactapi_header(app.current_request.to_dict(), env)

    @route("/header", authorize=False)
    def reactapi_route_header_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        No-env version of above /{env}/header route.
        """
        return app.core.reactapi_header(app.current_request.to_dict(), app.core.get_default_env())

    @route("/{env}/certificates", authorize=False)
    def reactapi_route_certificates(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        Returns SSL certificate info about this Foursight instance and its associated Portal instance.
        """
        ignored(env)
        return ReactRoutes.reactapi_route_certificates_noenv()

    @route("/certificates", authorize=False)
    def reactapi_route_certificates_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        Returns SSL certificate info about this Foursight instance and its associated Portal instance.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_certificates(request_dict, args)

    @route("/{env}/portal_access_key", authorize=False)
    def reactapi_route_portal_access_key(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        Returns Portal access key info about this Foursight instance.
        """
        ignored(env)
        return ReactRoutes.reactapi_route_portal_access_key_noenv()

    @route("/portal_access_key", authorize=False)
    def reactapi_route_portal_access_key_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        Returns Portal access key info about this Foursight instance.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_portal_access_key(request_dict, args)

    @route("/{env}/elasticsearch", authorize=False)
    def reactapi_route_elasticsearch_noenv(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        Pings the Elasticsearch associated with this Foursight instance and returns related info.
        """
        ignored(env)
        return ReactRoutes.reactapi_route_elasticsearch_noenv()

    @route("/elasticsearch", authorize=False)
    def reactapi_route_elasticsearch_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        """
        Note that this in an UNPROTECTED route.
        Pings the Elasticsearch associated with this Foursight instance and returns related info.
        """
        return app.core.reactapi_elasticsearch()

    # ----------------------------------------------------------------------------------------------
    # Foursight React API routes PROTECTED by authorization/authentication.
    # ----------------------------------------------------------------------------------------------

    @route("/{env}/info", authorize=True)
    def reactapi_route_info(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns sundry info about the app.
        """
        return app.core.reactapi_info(app.current_request.to_dict(), env)

    @route("/{env}/ecosystems", authorize=True)
    def reactapi_route_ecosystems(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns sundry info about the ecosystems.
        """
        return app.core.reactapi_ecosystems(app.current_request.to_dict(), env)

    @route("/ecosystems", authorize=True)
    def reactapi_route_ecosystems_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns sundry info about the ecosystems.
        """
        return app.core.reactapi_ecosystems(app.current_request.to_dict(), app.core.get_default_env())

    @route("/{env}/users", methods=["GET", "POST"], authorize=True)
    def reactapi_route_users_get_or_post(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        GET:  Returns the (paged) list of all defined users.
              Optional arguments (args) for the request are any of:
              - search: to search for the specified value.
              - limit: to limit the results to the specified number.
              - offset: to skip past the first specified number of results.
              - sort: to sort by the specified field name (optionally suffixed with .asc which is default or .desc);
                      default is email.asc.
              - raw: if true then returns the raw format of the data.
        POST: Creates a new user described by the given data;
              must contain: email, first_name, last_name.

        Note that you cannot have more than one route with the same path even
        if the methods are different; rather you must bundle them together and
        distinguish between which method is used programmatically as we do here.
        """
        if app.current_request.method == "GET":
            request_dict = app.current_request.to_dict()
            args = get_request_args(request_dict)
            return app.core.reactapi_get_users(request_dict, env, args)
        elif app.current_request.method == "POST":
            user = get_request_body(app.current_request)
            return app.core.reactapi_post_user(app.current_request.to_dict(), env, user=user)
        else:
            return app.core.create_forbidden_response()

    @route("/{env}/users/{uuid}", methods=["GET", "PATCH", "DELETE"], authorize=True)
    def reactapi_route_user_get_or_patch_or_delete(env: str, uuid: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        GET:    Returns detailed info for the user identified by the given uuid (may also be email).
                Optional arguments (args) for the request are any of:
                - raw: if true then returns the raw format of the data.
        PATCH:  Updates the user identified by the given uuid with the given user data;
                must contain: email, first_name, last_name.
        DELETE: Deletes the user identified by the given uuid.

        Note that you cannot have more than one route with the same path even
        if the methods are different; rather you must bundle them together and
        distinguish between which method is used programmatically as we do here.
        """
        request_dict = app.current_request.to_dict()
        if app.current_request.method == "GET":
            args = get_request_args(request_dict)
            return app.core.reactapi_get_user(request_dict, env, uuid, args)
        elif app.current_request.method == "PATCH":
            user = get_request_body(app.current_request)
            return app.core.reactapi_patch_user(request_dict, env, uuid=uuid, user=user)
        elif app.current_request.method == "DELETE":
            return app.core.reactapi_delete_user(request_dict, env, uuid=uuid)
        else:
            return app.core.create_forbidden_response()

    @route("/{env}/users/institutions", authorize=True)
    def reactapi_route_users_institutions(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the list of available user insitutions.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_users_institutions(request_dict, env, args)

    @route("/{env}/users/projects", authorize=True)
    def reactapi_route_users_projects(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the list of available user projects.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_users_projects(request_dict, env, args)

    @route("/{env}/users/awards", authorize=True)
    def reactapi_route_users_awards(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the list of available user awards.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_users_awards(request_dict, env, args)

    @route("/{env}/users/labs", authorize=True)
    def reactapi_route_users_labs(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the list of available user labs.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_users_labs(request_dict, env, args)

    @route("/{env}/users/consortia", authorize=True)
    def reactapi_route_users_consortia(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the list of available user consortia.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_users_consortia(request_dict, env, args)

    @route("/{env}/users/submission_centers", authorize=True)
    def reactapi_route_users_submission_centers(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the list of available user submission_centers.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_users_submission_centers(request_dict, env, args)

    @route("/{env}/users/roles", authorize=True)
    def reactapi_route_users_roles(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the list of available user roles.
        """
        request_dict = app.current_request.to_dict()
        return app.core.reactapi_users_roles(request_dict, env)

    @route("/{env}/users/schema", authorize=True)
    def reactapi_route_users_schema(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the user schema (JSON) from ElasticSearch.
        """
        request_dict = app.current_request.to_dict()
        return app.core.reactapi_users_schema(request_dict, env)

    @route("/{env}/users/statuses", authorize=True)
    def reactapi_route_users_statuses(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the possible user status.
        """
        request_dict = app.current_request.to_dict()
        return app.core.reactapi_users_statuses(request_dict, env)

    @route("/{env}/checks", authorize=True)
    def reactapi_route_checks_ungrouped(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns detailed info on all defined checks, NOT grouped by check group, as dictionary
        where each key name is the check name and its object contents contain the check details.
        For troubleshooting only.
        """
        return app.core.reactapi_checks_ungrouped(app.current_request.to_dict(), env)

    @route("/{env}/checks/grouped", authorize=True)
    def reactapi_route_checks_grouped(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns detailed info on all defined checks, grouped by check group, as a list where
        each item is an object containing the group name and list of check detail objects.
        """
        return app.core.reactapi_checks_grouped(app.current_request.to_dict(), env)

    @route("/{env}/checks/grouped/schedule", authorize=True)
    def reactapi_route_checks_grouped_by_schedule(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns detailed info on all defined checks, grouped by check group, as a list where each
        each item is an object containing the schedule namei and list list of check detail objecs.
        """
        return app.core.reactapi_checks_grouped_by_schedule(app.current_request.to_dict(), env)

    @route("/{env}/checks/{check}", authorize=True)
    def reactapi_route_checks_check(env: str, check: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the details of the given named checke as a dicitonary.
        """
        return app.core.reactapi_checks_check(app.current_request.to_dict(), env, check=check)

    @route("/{env}/checks/{check}/history", authorize=True)
    def reactapi_route_checks_history(env: str, check: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns detailed info on the run histories of the given check (paged).
        Optional arguments (args) for the request are any of:
        - limit: to limit the results to the specified number.
        - offset: to skip past the first specified number of results.
        - sort: to sort by the specified field name (optionally suffixed with .asc which is default or .desc);
                default value is timestamp.desc.
        """
        request_dict = app.current_request.to_dict()
        return app.core.reactapi_checks_history(request_dict, env, check=check, args=get_request_args(request_dict))

    @route("/{env}/checks/{check}/history/latest", authorize=True)
    def reactapi_route_checks_history_latest(env: str, check: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the latest (most recent) run result for the given check.
        """
        request_dict = app.current_request.to_dict()
        return app.core.reactapi_checks_history_latest(app.current_request.to_dict(), env, check=check)

    @route("/{env}/checks/{check}/history/{uuid}", authorize=True)
    def reactapi_route_checks_history_uuid(env: str, check: str, uuid: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the result of the given check result by its (name and) uuid.
        """
        return app.core.reactapi_checks_history_uuid(app.current_request.to_dict(), env, check=check, uuid=uuid)

    @route("/{env}/checks/history/recent", authorize=True)
    def reactapi_route_checks_history_recent(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns all "recent" check run history, across all checks.
        Paging is NOT currently supported, though the number of results can be limited
        via the limit argument, the default of which is 25; sorted by timestamp descending.
        Optional arguments (args) for the request are any of:
        - limit: to limit the results to the specified number; default is 25;
                 no additionl paging related functionality is currently supported.
        """
        request_dict = app.current_request.to_dict()
        return app.core.reactapi_checks_history_recent(request_dict, env, args=get_request_args(request_dict))

    @route("/{env}/checks/{check}/run", authorize=True)
    def reactapi_route_checks_run(env: str, check: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Kicks off a run of the given check.
        Arguments (args) for the request are any of:
        - args: Base-64 encode JSON object containing fields/values appropriate for the check run. 
        """
        request_dict = app.current_request.to_dict()
        return app.core.reactapi_checks_run(request_dict, env, check=check, args=get_request_arg(request_dict, "args"))

    @route("/{env}/action/{action}/run", authorize=True)
    def reactapi_route_action_run(env: str, action: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Kicks off a run of the given check.
        Arguments (args) for the request are any of:
        - args: Base-64 encode JSON object containing fields/values appropriate for the action run. 
        """
        request_dict = app.current_request.to_dict()
        return app.core.reactapi_action_run(request_dict, env, action=action, args=get_request_arg(request_dict, "args"))

    @route("/{env}/checks_status", authorize=True)
    def reactapi_route_checks_status(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on currently running/queueued checks.
        """
        return app.core.reactapi_checks_status(app.current_request.to_dict(), env)

    @route("/{env}/checks_raw", authorize=True)
    def reactapi_route_checks_raw(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns the contents of the raw check_setup.json file.
        """
        return app.core.reactapi_checks_raw(app.current_request.to_dict(), env)

    @route("/{env}/checks_registry", authorize=True)
    def reactapi_route_checks_registry(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns detailed registered checks functions.
        """
        return app.core.reactapi_checks_registry(app.current_request.to_dict(), env)

    @route("/{env}/checks_validation", authorize=True)
    def reactapi_route_checks_validation(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns information about any problems with the checks setup.
        """
        return app.core.reactapi_checks_validation(app.current_request.to_dict(), env)

    @route("/{env}/lambdas", authorize=True)
    def reactapi_route_lambdas(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns detailed info on defined lambdas.
        """
        return app.core.reactapi_lambdas(app.current_request.to_dict(), env)

    @route("/{env}/gac/{env_compare}", authorize=True)
    def reactapi_route_gac_compare(env: str, env_compare: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Compares and returns diffs for values in the given two GACs.
        """
        return app.core.reactapi_gac_compare(app.current_request.to_dict(), env, env_compare=env_compare)

    @route("/{env}/aws/s3/buckets", authorize=True)
    def reactapi_route_aws_s3_buckets(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Return the list of all AWS S3 bucket names for the current AWS environment.
        """
        return app.core.reactapi_aws_s3_buckets(app.current_request.to_dict(), env)

    @route("/{env}/aws/s3/buckets/{bucket}", authorize=True)
    def reactapi_route_aws_s3_buckets_keys(env: str, bucket: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Return the list of AWS S3 bucket key names in the given bucket for the current AWS environment.
        """
        return app.core.reactapi_aws_s3_buckets_keys(app.current_request.to_dict(), env, bucket=bucket)

    @route("/{env}/aws/s3/buckets/{bucket}/{key}", authorize=True)
    def reactapi_route_aws_s3_buckets_key_contents(env: str, bucket: str, key: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Return the content of the given AWS S3 bucket key in the given bucket for the current AWS environment.
        """
        return app.core.reactapi_aws_s3_buckets_key_contents(app.current_request.to_dict(), env, bucket=bucket, key=key)

    @route("/{env}/accounts", authorize=True)
    def reactapi_route_accounts(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on known accounts/environments as defined in an accounts.json file if present.
        """
        return app.core.reactapi_accounts(app.current_request.to_dict(), env)

    @route("/{env}/accounts/{name}", authorize=True)
    def reactapi_route_account(env: str, name: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on known accounts/environments as defined in an accounts.json file if present.
        Here are the pertinent data points returns by this endpoint, none of them secret:
        - Various version info (Python packages, Elasticsearch)
        - Foursight URL
        - Foursight deployed timestamp
        - Foursight stage name
        - Foursight current/default environment name (not all known environments but a count of them)
        - Global application configuration name i.e. identity (also in Portal health URL output)
        - S3 global environment bucket name
        - S3 Bucket organization name
        - AWS account ID and alias
        - S3 encryption ID (also in Portal health URL output)
        - Auth0 Client ID
        - reCAPTCHA ID
        - Elasticsearch hostname
        - RDS database hostname
        - Portal URL
        - Portal deployed timestamp
        - Portal health URL output (verbatim)
        """
        return app.core.reactapi_account(app.current_request.to_dict(), env, name)

    @route("/{env}/aws/vpcs", authorize=True)
    def reactapi_route_aws_vpcs(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS VPCs.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_vpcs(request_dict, env, args=args)

    @route("/{env}/aws/vpcs/{vpc}", authorize=True)
    def reactapi_route_aws_vpc(env: str, vpc: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS VPC for the given VPC ID..
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_vpcs(request_dict, env, vpc, args=args)

    @route("/{env}/aws/subnets", authorize=True)
    def reactapi_route_aws_subnets(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS Subnets.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_subnets(request_dict, env, args=args)

    @route("/{env}/aws/subnets/{subnet}", authorize=True)
    def reactapi_route_aws_subnet(env: str, subnet: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS Subnet.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_subnets(request_dict, env, subnet, args=args)

    @route("/{env}/aws/security_groups", authorize=True)
    def reactapi_route_aws_security_groups(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS Security Groups.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_security_groups(request_dict, env, args=args)

    @route("/{env}/aws/security_groups/{security_group}", authorize=True)
    def reactapi_route_aws_security_group(env: str, security_group: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS Security Group.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_security_groups(request_dict, env, security_group, args=args)

    @route("/{env}/aws/security_group_rules/{security_group}", authorize=True)
    def reactapi_route_aws_security_group(env: str, security_group: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS Security Group Rules.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_security_group_rules(request_dict, env, security_group, args=args)

    @route("/{env}/aws/network", authorize=True)
    def reactapi_route_aws_network(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS VPCs, Subnets, and Security Groups, grouped by VPC.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_network(request_dict, env, args=args)

    @route("/{env}/aws/network/{network}", authorize=True)
    def reactapi_route_aws_network(env: str, network: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS VPCs, Subnets, and Security Groups, grouped by VPC.
        Optional arguments (args) for the request are any of:
        - raw: if true then returns the raw format of the data.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_aws_network(request_dict, env, network, args=args)

    @route("/{env}/aws/stacks", authorize=True)
    def reactapi_route_aws_stacks(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info on AWS Stacks.
        """
        return app.core.reactapi_aws_stacks(app.current_request.to_dict(), env)

    @route("/{env}/aws/stacks/{stack}", authorize=True)
    def reactapi_route_aws_stack(env: str, stack: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns info for the given/named AWS Stack.
        """
        return app.core.reactapi_aws_stack(app.current_request.to_dict(), env, stack)

    @route("/{env}/aws/stacks/{stack}/outputs", authorize=True)
    def reactapi_route_aws_stack_outputs(env: str, stack: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns outputs for the given/named AWS Stack.
        """
        return app.core.reactapi_aws_stack_outputs(app.current_request.to_dict(), env, stack)

    @route("/{env}/aws/stacks/{stack}/parameters", authorize=True)
    def reactapi_route_aws_stack_parameters(env: str, stack: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns parameters for the given/named AWS Stack.
        """
        return app.core.reactapi_aws_stack_parameters(app.current_request.to_dict(), env, stack)

    @route("/{env}/aws/stacks/{stack}/resources", authorize=True)
    def reactapi_route_aws_stack_resources(env: str, stack: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns resources for the given/named AWS Stack.
        """
        return app.core.reactapi_aws_stack_resources(app.current_request.to_dict(), env, stack)

    @route("/{env}/aws/stacks/{stack}/template", authorize=True)
    def reactapi_route_aws_stack_template(env: str, stack: str) -> Response:  # noqa: implicit @staticmethod via @route
        """
        Returns template (as a string) for the given/named AWS Stack.
        """
        return app.core.reactapi_aws_stack_template(app.current_request.to_dict(), env, stack)

    @route("/__functioncache__", authorize=True)
    def reactapi_route_function_cache() -> Response:  # noqa: implicit @staticmethod via @route
        """
        For troubleshooting only. Returns function cache info.
        """
        return app.core.reactapi_function_cache(app.current_request.to_dict())

    @route("/__functioncacheclear__", authorize=True)
    def reactapi_route_function_cache() -> Response:  # noqa: implicit @staticmethod via @route
        """
        For troubleshooting only. Clears function cache.
        """
        request_dict = app.current_request.to_dict()
        args = get_request_args(request_dict)
        return app.core.reactapi_function_cache_clear(request_dict, args)

    @route("/__reloadlambda__", authorize=True)
    def reactapi_route_reload_lambda() -> Response:  # noqa: implicit @staticmethod via @route
        """
        For troubleshooting only. Reload the lambda code.
        """
        return app.core.reactapi_reload_lambda(app.current_request.to_dict())

    @route("/__testsize__/{n}", authorize=True)
    def reactapi_route_testsize(n: int) -> Response:  # noqa: implicit @staticmethod via @route
        """
        For troubleshooting only. Test response size capabilities of AWS Lamdas.
        """
        return app.core.reactapi_testsize(n)

    @route("/{env}/accounts_file", methods=["POST"], authorize=True)
    def reactapi_route_accounts_file_upload(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        ignored(env)
        return ReactRoutes.reactapi_route_accounts_file_upload_noenv()

    @route("/accounts_file", methods=["POST"], authorize=True)
    def reactapi_route_accounts_file_upload_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        accounts_file_data = get_request_body(app.current_request)
        return app.core.reactapi_accounts_file_upload(accounts_file_data)

    @route("/{env}/accounts_file", authorize=True)
    def reactapi_route_accounts_file_download(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        ignored(env)
        return ReactRoutes.reactapi_route_accounts_file_download_noenv()

    @route("/accounts_file", authorize=True)
    def reactapi_route_accounts_file_download_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.reactapi_accounts_file_download()

    @route("/{env}/aws/secrets/{secrets_name}", authorize=True)
    def reactapi_route_aws_secrets(env: str, secrets_name: str) -> Response:  # noqa: implicit @staticmethod via @route
        ignored(env)
        return ReactRoutes.reactapi_route_aws_secrets_noenv(secrets_name)

    @route("/aws/secrets/{secrets_name}", authorize=True)
    def reactapi_route_aws_secrets_noenv(secrets_name: str) -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.reactapi_aws_secrets(secrets_name)

    @route("/{env}/aws/secrets", authorize=True)
    def reactapi_route_aws_secret_names(env: str) -> Response:  # noqa: implicit @staticmethod via @route
        ignored(env)
        return ReactRoutes.reactapi_route_aws_secret_names_noenv()

    @route("/aws/secrets", authorize=True)
    def reactapi_route_aws_secret_names_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.reactapi_aws_secret_names()

    @route("/aws/ecs/clusters", authorize=True)
    def reactapi_route_aws_ecs_clusters() -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.reactapi_aws_ecs_clusters()

    @route("/aws/ecs/clusters/{cluster_arn}", authorize=True)
    def reactapi_route_aws_ecs_cluster(cluster_arn: str) -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.reactapi_aws_ecs_cluster(cluster_arn=cluster_arn)

    @route("/aws/ecs/tasks", authorize=True)
    def reactapi_route_aws_ecs_tasks() -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.reactapi_aws_ecs_task_arns(latest=False)

    @route("/aws/ecs/tasks/{task_definition_arn}", authorize=True)
    def reactapi_route_aws_ecs_task(task_definition_arn: str) -> Response:  # noqa: implicit @staticmethod via @route
        if task_definition_arn.lower() == "latest":
            return app.core.reactapi_aws_ecs_task_arns(latest=True)
        elif task_definition_arn.lower() == "details":
            return app.core.reactapi_aws_ecs_tasks(latest=False)
        return app.core.reactapi_aws_ecs_task(task_definition_arn)

    @route("/aws/ecs/tasks/latest/details", authorize=True)
    def reactapi_route_aws_ecs_task() -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.reactapi_aws_ecs_tasks(latest=True)

    # ----------------------------------------------------------------------------------------------
    # Foursight React UI (static file) routes, serving the HTML/CSS/JavaScript/React files.
    # Note that ALL of these are UNPROTECTED routes.
    # ----------------------------------------------------------------------------------------------

    # TODO: See if there is a better way to deal with variadic paths.
    # TODO: Maybe end up serving these from S3, for more security, and smaller Chalice package size.

    @route(root=True)
    def reactui_route_root() -> Response:  # noqa: implicit @staticmethod via @route
        return route_root()

    @route("/", static=True, authorize=False)
    def reactui_route_static_file_noenv() -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.react_serve_static_file(app.core.get_default_env(), [])

    @route("/{env}", static=True, authorize=False)
    def reactui_route_0(env) -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.react_serve_static_file(env, [])

    @route("/{env}/{path1}", static=True, authorize=False)
    def reactui_route_1(env, path1) -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.react_serve_static_file(env, [path1])

    @route("/{env}/{path1}/{path2}", static=True, authorize=False)
    def reactui_route_2(env, path1, path2) -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.react_serve_static_file(env, [path1, path2])

    @route("/{env}/{path1}/{path2}/{path3}", static=True, authorize=False)
    def reactui_route_3(env, path1, path2, path3) -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.react_serve_static_file(env, [path1, path2, path3])

    @route("/{env}/{path1}/{path2}/{path3}/{path4}", static=True, authorize=False)
    def reactui_route_4(env, path1, path2, path3, path4) -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.react_serve_static_file(env, [path1, path2, path3, path4])

    @route("/{env}/{path1}/{path2}/{path3}/{path4}/{path5}", static=True, authorize=False)
    def reactui_route_5(env, path1, path2, path3, path4, path5) -> Response:  # noqa: implicit @staticmethod via @route
        return app.core.react_serve_static_file(env, [path1, path2, path3, path4, path5])
